import sys
from . import mytypes
from . import glb

# Record some info for cpp functions

# {func_name => [return_type, cpp_module]}
# '' means built-in functions

# Fields (default value):
# 1. return type (mytypes.void)
# 2. cpp module ('')
# 3. number of versions (1)
# 4. a bit vector: availability in -O0,1,2 (4)
#      __,__,__
#      O2 O1 O0
#      4: only available in -O2
 
float64_reduction = (mytypes.float64, 'reduction')
int64List_reduction = (mytypes.int64_list, 'reduction')
bool_reduction = (mytypes.bool, 'reduction')
# TODO
# means return type should be the element type pf the first argument
# arg0ElemType_reduction = (mytypes.float64, 'reduction')
# arg0Type_elemwise = (mytypes.float64_ndarray, 'elemwise')

def func_zeros_ret_type(arg_types):
    ndim = -1
    shape_ty = arg_types[0]
    if mytypes.is_list(shape_ty) \
       or mytypes.is_tuple(shape_ty):
        ndim = shape_ty.get_init_size()
    if mytypes.is_int(shape_ty):
        ndim = -1

    # -- if data type not specified
    if len(arg_types) == 1:
        return mytypes.NpArray(mytypes.float64, ndim)
    # -- if data type specified    
    elif len(arg_types) == 2:
        return mytypes.NpArray(arg_types[1], ndim)
    else:
        assert False

def func_zeros_2d_ret_type(arg_types):
    if len(arg_types) == 1:
        return mytypes.float64_2darray
    elif len(arg_types) == 2:
        return mytypes.NpArray(arg_types[1], 2)
    else:
        assert False

def func_sum_ret_type(arg_types):
    if len(arg_types) == 1:
        return arg_types[0].get_elt_type() # Is this right Jun? => Right
    elif len(arg_types) == 2 and isinstance(arg_types[0], mytypes.NpArray):
        return arg_types[0]
    else:
        assert False

def array_or_scalar(arg_types, elt_type):
    if isinstance(arg_types[0], mytypes.NpArray):
        return mytypes.NpArray(elt_type)
    elif len(arg_types) > 1 and isinstance(arg_types[1], mytypes.NpArray):
        return mytypes.NpArray(elt_type)
    else:
        return elt_type
    
funcinfo = {
    'print': (mytypes.void, '', 1, 5),
    'append': (mytypes.void, ''),
    'fill': (mytypes.void, ''),
    'len': (mytypes.int64, '', 1, 5),
    'boolean': [mytypes.bool, ''],
    'int32': [mytypes.int32, ''],
    'int64': [mytypes.int64, ''],
    'float32': [mytypes.float32, ''],
    'float64': [mytypes.float64, ''],
    'time': [mytypes.float64, ''],
    'range': [mytypes.int32_list,
              'NpArray', 1, 5],
    'zeros': (func_zeros_ret_type,
              'NpArray'),
    'zeros_2d': (func_zeros_2d_ret_type,
              'NpArray'),
    'empty': 'zeros', # same as zeros
    'arange': 'empty',
    'copy': 'empty',
    'numpy_random_rand': 'zeros',
    'empty_2d': 'zeros_2d',
    'empty_like': (lambda arg_types: arg_types[0],
              'NpArray'),
    'get_row': (lambda arg_types: arg_types[0],
              'NpArray'),
    'get_row_ptr': (lambda arg_types: arg_types[0],
              'NpArray'),
    'get_col': 'get_row',

    'set_row': (mytypes.void, 'NpArray'),
    'set_col': (mytypes.void, 'NpArray'),
    'sum_rows': (lambda arg_types: arg_types[0], 'NpArray'),
    'plus_eq': (mytypes.void, 'NpArray'),
    'minus_eq': (mytypes.void, 'NpArray'),
    'shape': [mytypes.int64,
              'NpArray'],
    'stride': 'shape',
    'arraysub': [mytypes.void,
                 'NpArray'],
    'dsyrk': [mytypes.float64_ndarray,
                   'lib'],
    'omp': [mytypes.void,
            'omp'],
    # Reduction (TODO: add implementation versions)
    'sum': (lambda arg_types: func_sum_ret_type(arg_types), 'reduction'),
    'prod': 'sum', # means same as above
    'min': 'sum',
    'max': 'sum',
    'argmin': int64List_reduction,
    'argmax': int64List_reduction,
    'any': bool_reduction,
    'all': 'any',
    'allclose': 'any',
    'where': [mytypes.int32_1darray, 'reduction'],
    'where1': [mytypes.int32_1darray, 'reduction'],
    # Element-wise unary (TODO: add implementation versions)
    'abs': (lambda arg_types: array_or_scalar(arg_types, mytypes.float64), 'elemwise'),
    'minus': (lambda arg_types: arg_types[0], 'elemwise'),
    'isnan': (lambda arg_types: array_or_scalar(arg_types, mytypes.bool), 'elemwise'),
    'isinf': 'isnan',
    'elemwise_not': 'isnan',
    'sqrt': (lambda arg_types: array_or_scalar(arg_types, mytypes.float64), 'elemwise'),
    'exp': 'sqrt',
    'cos': 'sqrt',
    'sin': 'sqrt',
    'tan': 'sqrt',
    'acos': 'sqrt',
    'asin': 'sqrt',
    'atan': 'sqrt',
    # Element-wise binary (TODO: add implementation versions)
    'add': [lambda arg_types: arg_types[0] if arg_types[0].equals(arg_types[1]) else mytypes.float64_ndarray,
            'elemwise'],
    'sub': 'add',
    'mul': 'add',
    '_sub': (mytypes.void, 'elemwise'),
    'maximum': (lambda arg_types: array_or_scalar(arg_types, mytypes.float64), 'elemwise'),
    'pow': (lambda arg_types: array_or_scalar(arg_types, mytypes.float64), 'elemwise'),
    'div': 'pow',
    'log': 'pow',
    'eq': [mytypes.bool_ndarray,
           'elemwise'],
    'neq': 'eq',
    'lt': 'eq',
    'gt': 'eq',
    'le': 'eq',
    'ge': 'eq',
    'logical_and' : 'eq',
    'logical_or' : 'eq',
    'logical_xor' : (lambda arg_types: array_or_scalar(arg_types, mytypes.bool), 'elemwise'),
    'logical_not' : 'isnan',
    'compatibility_check' : [mytypes.bool, 'elemwise'],
    # Matrix operations
    'transpose': [lambda arg_types: arg_types[0],
                  'matrixop'],
    'innerprod': [mytypes.float64,
                  'matrixop'],
    'innerprod1': [mytypes.float64,
                  'matrixop'],
    'matmult1': [mytypes.int32,
                  'matrixop'],
    'transpose1': [mytypes.int32,
                  'matrixop'],
    'matmult': [mytypes.float64_ndarray,
                'matrixop', 2],
    'syrk': 'matmult',
    'syr2k': 'matmult',
    'symm': 'matmult',
    'trmm': 'matmult',
    'lu': 'matmult',
    'ludcmp': 'matmult',
    'qr': 'matmult',
    'eig': 'matmult',
    'svd': 'matmult',
    'tril': 'matmult',
    'triu': 'matmult',
    'diag': 'matmult',
    'kron': 'matmult',
    'convolve': 'matmult',
    # Sparse matrix operations
    'empty_spm': [mytypes.float64_sparray,   # TODO: can be int and etc
                  'sparsemat'],
    'csr_to_spm': 'empty_spm',
    'arr_to_spm': 'empty_spm',
    'spm_to_csr': [mytypes.int32,
                   'sparsemat'],
    'spm_set_item': [mytypes.void,
                     'sparsemat'],
    'spm_set_item_unsafe': 'spm_set_item',
    'getval': [mytypes.float64,
               'sparsemat'],
    'getnnz': [mytypes.int32,
               'sparsemat'],
    'getcol': 'getnnz',
    'nnz': 'getnnz',
    'spm_add': [mytypes.float64_sparray,
                'sparsemat'],
    'spm_mul': 'spm_add',
    'spmm': 'spm_add',
    'spmv': [mytypes.float64_ndarray,
             'sparsemat'],
    'spmm_dense': [mytypes.float64_2darray,
             'sparsemat'],
    'sparse_diags': 'empty_spm',
    'spm_diags': 'empty_spm',
    'spm_mask': [mytypes.float64_sparray,
            'sparsemat'],
    'heapinit_empty': [mytypes.int32_int32_heap,
                 'heapq'],
    'heapinit': [mytypes.float32_heap,
                 'heapq'],
    'heappush': [mytypes.void,
                 'heapq'],
    'heappop': [mytypes.float32_list,
                'heapq'],
    'heappop1': [mytypes.void,
                'heapq'],
    'heappeek': [mytypes.float32,
                 'heapq'],
    'heapsize': [mytypes.int32,
                 'heapq'],

    'heap_get_key': [mytypes.int32,
                 'heapq'],
    'randint': [mytypes.int32, ''],
    'hex': [mytypes.string, ''],
    'hextoi': [mytypes.int32, ''],
    'stoi': [mytypes.int32, ''],
    'strtol': [mytypes.int32, ''],
}

# {cpp_library => [compiler_flags, linker_flags]}
# FIXME: dont think this is in use anymore
packageinfo = {
    'NpArray': [],
    'reduction': [],
    'elemwise': [],
    'matrixop': [],
    'sparsemat': [],
    'heapq': [],
    'omp': ['-fopenmp', '-fopenmp'],
    'lib': ['-I /usr/include/openblas',
            '-lblas'],
}


#used_packages = {'NpArray', 'reduction', 'elemwise', 'matrixop', 'sparsemat', 'heapq', 'omp'}
used_packages = ['shared/NpArray', 'py/reduction', 'py/elemwise',
                'py/matrixop', 'py/sparsemat', 'py/heapq']
#used_packages = {'NpArray'}


# global/built-in/library?
def is_global_func(func_name, module=''):
    # # TODO: to fix this
    return func_name in funcinfo
    # opt_bits = get_opt_level_bits(func_name)
    # is_available = 2**glb.args.opt_level & opt_bits
    # return func_name in funcinfo and is_available

def get_func_info(funcname, module='pydd'):
    '''
    A function is uniquely identiied by its module, name and signature.
    But due to the format of funcinfo table, we don't deal with argument
    signature here.
    '''    
    if module == 'pydd':
        if funcname in funcinfo:
            v = funcinfo[funcname]
            while isinstance(v, str):
                v = funcinfo[v]
            return v

            # if isinstance(v, str):                
            #     return get_func_info(v, module)
            # else:
            #     return v
        else:
            raise glb.UndefinedSymbolException(module, funcname)
    else:
        m_info = glb.get_imported_module_info(module)
        v = m_info.funcinfo[funcname]
        while isinstance(v, str):
            v = m_info.funcinfo[v]
        return v

# def get_ret_type(call_sig):
# #    print(vars(call_sig))
#     info = get_func_info(call_sig.funcname, call_sig.module)
#     if info and len(info) > 0:
#         ty = info[0]
#         if callable(ty):
#             return ty(call_sig.arg_types)
#         else:
#             return ty
#     else:
#         return mytypes.void

def get_type(name, module, arg_types=None):
    '''
    Get type of an external symbol via table lookup
    '''
    info = get_func_info(name, module)
    if info and len(info) > 0:
        ty = info[0]
        if callable(ty):
            return ty(arg_types)
        else:
            return ty
    else:
        return mytypes.void

def get_package(func_name):
    info = get_func_info(func_name)
    if info and len(info) > 1:
        return info[1]
    else:
        return ''

def get_version_count(func_name, module='pydd'):
    # # TODO: to fix this
    # return 1
    if module == 'std':
        return 1
    info = get_func_info(func_name, module)
    if info and len(info) > 2:
        return info[2]
    return 1

def get_opt_level_bits(func_name):
    info = get_func_info(func_name)
    if info and len(info) > 3:
        return info[3]
    return 4

def is_in_O0(func_name):
    info = get_func_info(func_name)
    if info and len(info) > 3:
        return info[3] & 1
    return False

def is_in_O1(func_name):
    info = get_func_info(func_name)
    if info and len(info) > 3:
        return info[3] & 2
    return False

def is_in_O2(func_name):
    info = get_func_info(func_name)
    if info and len(info) > 3:
        return info[3] & 4
    return True

def add_function_module(func_name):
    return
    global used_packages
    p = get_package(func_name)
    if p:
        used_packages.add(p)

# def register_lib_call(func_name):
#     global used_packages
#     used_packages.add(get_package(func_name))

def get_header_files():
    s = []

    # packages = ['shared/NpArray', 'py/reduction', 'py/elemwise',
    #             'py/matrixop', 'py/sparsemat', 'py/heapq', 'shared/omp']
    packages = used_packages

    #for p in used_packages:
    for p in packages:
        if p.startswith('shared/'):
            s.append(p+'.hpp')
        elif p.startswith('py/'):
            if glb.args.host == 'py':
                s.append(p+'.hpp')

    for p in glb.cpp_module.imports:
        if glb.is_imported_module_standard(p):
            s.append(p)
        else:
            s.append('%s/%s.hpp' % (p, p))
        
    return s

def get_compiler_flags():
    anaconda_path = sys.executable.replace('bin/python', '')
    s = ['-m64', '-I'+anaconda_path+'/include']
    # for p in used_packages:
    #     flags = packageinfo[p]
    #     if len(flags) > 0:
    #         s.append(flags[0])
    return s

def get_linker_flags():
    s = []

    if glb.args.blas:
        # Anaconda should already have these
        # s += ['-liomp5', '-lmkl_core', '-lmkl_intel_thread',
        #       '-lmkl_intel_lp64', '-pthread']
        
        #from pathlib import Path
        #s += f"-m64 -L{Path.home()}/anaconda3/lib -Wl,--no-as-needed -lmkl_intel_ilp64 -lmkl_gnu_thread -lmkl_core -lgomp -lpthread -lm -ldl".split()
        anaconda_path = sys.executable.replace('bin/python', '')
        # remove "-Wl,--no-as-needed" to make it work on Mac
        from sys import platform
        if platform == "linux" or platform == "linux2":
            s += f"-m64 -L{anaconda_path}/lib -Wl,--no-as-needed -lmkl_intel_ilp64 -lmkl_gnu_thread -lmkl_core -lgomp -lpthread -lm -ldl".split()
        elif platform == "darwin":
            # OS X
            s += f"-m64 -lopenblas -lgomp -lpthread -lm -ldl".split()
        elif platform == "win32":
            # Windows...
            raise Exception('Windows is to be supported!')
    
    # for p in used_packages:
    #     flags = packageinfo[p]
    #     if len(flags) > 1:
    #         s.append(flags[1])
    return s



import ast
from . import libfuncs
from . import glb
from . import scalarpriv
from . import licmarray
from . import licmast
from . import opexpansion
from . import arrayopt

class TypedTransformer(ast.NodeVisitor):
    def visit(self, tree):
        # Todo: invoke operator expansion
        #  - E.g., convert "A = B * C" into "A = mul(B, C)" when B and/or C has array type.
        #  - Create a new file for this functionality (similar to scalarpriv/loopinvariant below)
        opexpansion.OperatorExpansion().visit(tree)

        # Loop Invariant Code Motion for AST-level expressions (general)
        if 'licm_ast' in glb.options and glb.options['licm_ast']:
            licmast.LicmAstOptimization().visit(tree)

        # Sparse/dense array optimizations
        if ('sparse_opt' in glb.options and glb.options['sparse_opt']
            or 'dense_opt' in glb.options and glb.options['dense_opt']):
            arrayopt.ArrayOptimization().visit(tree)

        # Privatization for parallel loops (pfor)
        scalarpriv.ScalarPrivatization().visit(tree)

        # Loop Invariant Code Motion for array references
        if 'licm_array' in glb.options and glb.options['licm_array']:
            licmarray.LicmArrayAnalysis().visit(tree)



import ast

class ReturnExprToStmt(ast.NodeTransformer):
    def visit_Return(self, node):
        if not isinstance(node.value, ast.Name):
            assign = ast.Assign(targets = [ast.Name(id = '__ret', ctx = ast.Store())], value = node.value, lineno = node.lineno, col_offset = node.col_offset)
            node.value = ast.Name(id = '__ret', ctx = ast.Load())
            return [assign] + [node]
        else:
            return node



#!/usr/bin/env python3

from . import launcher
from . import glb
import cppimport
import sys
import os, errno
import io
import subprocess

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

def get_indented_as_string(filename):
    try:
        output = subprocess.check_output(["clang-format", filename])
        return output
    except OSError as e:
        if e.errno == errno.ENOENT:
            pass
            # handle file not found error.
        else:
            pass
            # Something else went wrong while trying to run `wget`
    

            

def create_indented_cpp(filename):
    '''
    Create a stripped/valid cpp file from the generated kernel.cpp
    '''
    
    header = ''
    body = ''
    footer = ''
    state = 0
    for line in open(filename):
        if state == 0:
            header += line
        elif state == 1:
            if not line.startswith('PYBIND11_PLUGIN('):
                body += line
        elif state == 2:
            footer += line
            
        if state == 0 and line.startswith('namespace py'):
            state = 1

        if state == 1 and line.startswith('PYBIND11_PLUGIN('):
            state = 2
            footer += line


    newname = filename.replace('.cpp', '_indented.cpp')
    newfile = open(newname, 'w')
    newfile.write(body)
    newfile.close()
    output = ''
    try:
        body = subprocess.check_output(["clang-format", newname]).decode("utf-8")
    except OSError as e:
        if e.errno == errno.ENOENT:
            print("[Tips]: installing clang-format (apt install clang-format) will indent the generated cpp file")
            # handle file not found error.
        else:
            print("cpp file formatting failed")
            
            # Something else went wrong while trying to run `wget`
            
    newfile = open(newname, 'w')
    newfile.write(header)
    newfile.write(body)
    newfile.write(footer)
    newfile.close()
    os.rename(newname, filename)

def compile(file):
    sys.argv[0] = file
    main()

def main():
    glb.init()
    os.chdir(glb.get_filepath())

    # Remove old files
    module = glb.get_module_name()
    # subprocess.call(['find', '.', '-name',
    #                  '%s.cpython*.so' % module,
    #                  '-or', '-name', '%s.cpp' % module,
    #                  '-delete'])   

    # Launch compiler
    launcher.main()


    if glb.args.host == 'cpp':
        sys.exit(0)

    if glb.args.opt_level == 2:
        if glb.args.no_compile:
            sys.exit(0)
    ### Doesn't seem to a general solution, commented out for now    
        #   if sys.platform.startswith('darwin'):
        #        os.environ["MACOSX_DEPLOYMENT_TARGET"] = "10.9"
        #        os.environ["CC"] = "clang"
        #        os.environ["CXX"] = "clang++"

        sys.path.append(os.getcwd())

        import time
        time_start = time.time()
        gcc_err = open(module+'_gcc.err', 'w')
        gcc_out = open(module+'_gcc.out', 'w')
        ret = subprocess.call(
            [sys.executable, glb.get_compiler_path()+"/runcppimport.py", module],
            stderr=gcc_err, stdout=gcc_out)

        time_end = time.time()
        print(time_end - time_start)
        if not glb.args.file.endswith('.cpp'):
            create_indented_cpp(module+'.cpp')
        if ret != 0:
            print("Compilation failed, gcc return code: %d."%ret)
            print("Check `./%s_gcc.err` for details." % module)

            if glb.args.verbose:
                for line in open(module+'_gcc.err', 'r'):
                    if 'error:' in line:
                        print(bcolors.FAIL+line+bcolors.ENDC)
                    else:
                        print(line)

            sys.exit(ret)            

    
if __name__ == "__main__":
    main()


import sys
import ast
from pprint import pprint
import traceback

from . import mytypes
from .codegen import ModuleGen
from . import glb
from . import utils
from . import libfuncs
from .symboltable import symtab
from .glb import UnhandledNodeException
from .glb import dump
from .symboltable import CallSig

class TypeInferer(ast.NodeVisitor):
    def __init__(self):
        self.stats = {"import": [], "from": []}
        self.cur_func = None
        
    def visit_Module(self, N):
        self.generic_visit(N)

    def get_cur_func(self):
        return self.cur_func
        
    def visit_FunctionDef(self, N: ast.FunctionDef):
        self.typemap = N.localtypes
        self.cur_func = N
        self.cur_func.has_pfor = False
       
        self.do_func_args(N)
        #self.gen_func_header(N)
        self.do_func_body(N)
        N.typemap = self.typemap
        #print(self.typemap)
        #self.generic_visit(node)

        glb.cpp_module.add_function(N)    
        
    def do_func_args(self, N: ast.FunctionDef):
        pendings = []
        for arg in N.args.args:
            if not arg.annotation:
                pendings.append(arg)
                continue
            ty = utils.get_annotation_type(arg.annotation)
            arg.type = ty
            
            for pending_arg in pendings:
                pending_arg.type = ty
                self.record_type(pending_arg)
                
            pendings.clear()
            self.record_type(arg)
        if len(pendings) > 0:
            print("Missing type annotations for argument ", end='')
            for arg in pendings:
                print('`%s`'%arg, end='')
            print()
            print('In function `%s`' % self.get_cur_func.name)    

    # def record_type(self, name, ty):
    #     self.typemap[name] = ty

    def record_type(self, N):
        # TODO: to check if an exsiting type is compatible with the new type
        if isinstance(N, ast.Name):
            self.typemap[N.id] = N.type
        elif isinstance(N, ast.arg):
            self.typemap[N.arg] = N.type
        else:
            glb.exit_on_unsupported_node(N)
            
            
    def do_func_body(self, N: ast.FunctionDef):
        for stmt in N.body:
            if self.is_str_stmt(stmt):
                continue
            self.do_node(stmt)

    def is_str_stmt(self, N):
        if isinstance(N, ast.Expr):
            if isinstance(N.value, ast.Str):
                return True
        return False    

    def do_node(self, N):
        if isinstance(N, ast.Assign):
            self.do_Assign(N)
        elif isinstance(N, ast.AnnAssign):
            self.do_AnnAssign(N)
        elif isinstance(N, ast.AugAssign):
            self.do_AugAssign(N)
        elif isinstance(N, ast.Num):
            self.do_Num(N)
        elif isinstance(N, ast.Name):
            self.do_name(N)
        elif isinstance(N, ast.For):
            self.do_For(N)
        elif isinstance(N, ast.While):
            self.do_While(N)
        elif isinstance(N, ast.If):
            self.do_If(N)
        elif isinstance(N, ast.Return):
            self.do_Return(N)
        elif isinstance(N, ast.NameConstant):
            self.do_name_constant(N)
        elif isinstance(N, ast.Tuple):
            self.do_Tuple(N)
        elif isinstance(N, ast.List):
            self.do_List(N)
        elif isinstance(N, ast.UnaryOp):
            self.do_UnaryOp(N)    
        elif isinstance(N, ast.BinOp):
            self.do_BinOp(N)
        elif isinstance(N, ast.BoolOp):
            self.do_BoolOp(N)
        elif isinstance(N, ast.Compare):
            self.do_Compare(N)    
        elif isinstance(N, ast.Expr):
            self.do_Expr(N)
        elif isinstance(N, ast.Str):
            self.do_Str(N)
        elif isinstance(N, ast.Subscript):
            self.do_Subscript(N)
            
        elif isinstance(N, ast.Call):
            self.do_Call(N)
    
        else:
            pass
                
    # def get_type(self, N):
    #     if not hasattr(N, 'type'):
    #         self.do_node(N)
    #     return a.type

    def get_symbol_type(self, name: ast.Name):
        assert type(name) == ast.Name, ast.dump(name)
        if name.id in self.typemap:
            return self.typemap[name.id]
        else:
            glb.exit_on_node(name, 'typed unresolved for symbol')
            raise Exception("typed unresolved for symbol `" + name + '`')

    def do_BoolOp(self, N):
        N.type = mytypes.bool

    def do_Compare(self, N):
        N.type = mytypes.bool
        
    def do_UnaryOp(self, N):
        self.do_node(N.operand)
        N.type = N.operand.type

    def do_BinOp(self, N):
        self.do_node(N.right)
        self.do_node(N.left)
        lty = N.left.type
        rty = N.right.type
        # print(ast.dump(N))
        # print(lty)
        # print(rty)
        if mytypes.is_all_float(lty, rty) or mytypes.is_all_int(lty, rty):
            if lty.get_width() >= rty.get_width():
                N.type = lty
            else:
                N.type = rty
            return
        
        
        if not N.left.type.can_broadcast_with(N.right.type):
            msg = "type unmatch in binary operator %s\n" % N.op
            msg += "left (%s): %s\n" % (N.left.type, ast.dump(N.left))
            msg += "right (%s): %s\n" % (N.right.type, ast.dump(N.right))
            if glb.args.verbose:
                print("nn")
                dump(N.left)
                dump(N.right)
            glb.exit_on_node(N, msg)

        # The list type should dominate if it exists in the BinOp
        # a = [1, 2, 3] * 4  <- a should have type array<int>
        # Otherwise, just take either of the two types, since they are the same
        N.type = N.left.type if hasattr(N.left.type, 'etype') else N.right.type

    def do_Str(self, N):
        N.type = mytypes.string

    def verify_array_dim(self, N: ast.Subscript):
        if N.value.id in self.typemap:
            ty = self.get_symbol_type(N.value)
            if (isinstance(ty, mytypes.NpArray) and ty.ndim != -1):
                print(ast.dump(N))
                if isinstance(N.slice, ast.Tuple):
                #if isinstance(N.slice, ast.Tuple):
                    ndim_found = len(N.slice.elts)
                else:
                    ndim_found = 1
                if ty.ndim != ndim_found:
                    msg = 'annotated # array dims = %d while found = %d' % (ty.ndim, ndim_found)
                    #glb.exit_on_node(N, msg)
                    
                    raise Exception(msg)

    def do_Subscript(self, N):
        if type(N.value) == ast.Attribute and N.value.attr == 'shape':
            glb.exit_on_node(N, "Please use `shape()` instead of `shape[]`")
            assert False, "haha"

        basetype = self.get_symbol_type(N.value)
        N.type = basetype.get_elt_type()
        if isinstance(basetype, mytypes.DictType):
            N.type = basetype.value

        libfuncs.add_function_module('arraysub')  # Arrays can be accessed without range/shape
        self.verify_array_dim(N)

    def do_Expr(self, N: ast.Expr):
        if isinstance(N.value, ast.Call):
            self.do_Call(N.value)
        else:
            glb.exit_on_unsupported_node(N)

    def do_Call(self, N: ast.Call):
        # eg: Call(func=Name(id='print', ctx=Load()), args=[Name(id='i', ctx=Load())], keywords=[])
        # Call(func=Attribute(value=Name(id='ls2', ctx=Load()), attr='append', ctx=Load()), args=[Name(id='j', ctx=Load())], keywords=[])
        # todo
        arg_types = []
        for arg in N.args:
            self.do_node(arg)
            if not hasattr(arg, "type"):
                glb.exit_on_node(N, "Node type unknown")
            arg_types.append(arg.type)
        for key in N.keywords:
            self.do_node(key)

        callSig = None
        module = ''
        funcname = ''
        try:
            if isinstance(N.func, ast.Name):
                # Either should be a user-defined function or a global function
                funcname = N.func.id

                if funcname.startswith('_'):
                    # functions like _add/_sub have no meaningful return value for now
                    return
                
                if not glb.is_user_defined_func(funcname):
                    module = 'pydd'
                    libfuncs.add_function_module(funcname)
            elif isinstance(N.func, ast.Attribute):
                # Should be a module.function
                f = N.func        
                if not glb.cpp_module.is_module(f.value.id):
                    glb.exit_on_unsupported_node(N)
                else:
                    module = f.value.id
                    funcname = f.attr
                    N.is_module_call = True
                    
            #callSig = CallSig(module, funcname, arg_types)

            N.type = symtab.get_type(funcname, module, arg_types)
            # if funcname == 'empty':
            #     print(arg_types)
            #     print("empty call type:", N.type.get_dimension())
            #     sys.exit(1)

        except glb.UndefinedSymbolException as e:
            glb.exit_on_node(N, str(e))

            
    def eval_type_comment(self, s: str):
        print("got type:", s)
        if s == '[int]':
            return mytypes.int64_list
        if s == 'int' or s == 'int32':
            return mytypes.int32
        elif s == 'int64':
            return mytypes.int64
        elif s == 'double' or s == 'float64':
            return mytypes.float64
        elif s == 'float' or s == 'float32':
            return mytypes.float32
        else:
            return Exception('Unrecognized type comment: ' + s)

    def do_AnnAssign(self, N: ast.AnnAssign):
        ty = utils.get_annotation_type(N.annotation)
        N.target.type = N.value.type = ty
        self.record_type(N.target)
        N.targets = [N.target]
        self.do_Assign_common(N)
           
    def do_AugAssign(self, N: ast.AugAssign):
        self.do_node(N.value)
        N.targets = [N.target]
        self.do_Assign_common(N)

    def do_Assign(self, N: ast.Assign):
        
        if N.type_comment:
            N.value.type = self.eval_type_comment(N.type_comment)
        else:
            #print("do_Assign", ast.dump(N))
            self.do_node(N.value)
            ty = N.value.type
        self.do_Assign_common(N) 
        

    def do_Assign_common(self, N):
        for target in N.targets:
            # Assign to one var
            if isinstance(target, ast.Name):
                if hasattr(target, 'type'):
                    continue
                target.type = N.value.type
                self.record_type(target)
            # Assign to multiple vars    
            elif isinstance(target, ast.Tuple):
                var_tp = target.elts
                value_tp = N.value.elts
                assert len(var_tp) == len(value_tp)
                for i in range(len(var_tp)):
                    var_tp[i].type = value_tp[i].type
                    self.record_type(var_tp[i])
            # TODO: var = BinOp
            
            if isinstance(target, ast.Subscript):
                self.verify_array_dim(target)
            


    def do_name(self, N):
        N.type = self.get_symbol_type(N)
        
    def do_tuple_assign(self, N: ast.Assign):
        value_tp = N.value.elts
        var_tp = N.targets[0].elts
        if len(value_tp) != len(var_tp):
            raise SyntaxError("# of elements in the left and right don't match in assignment") 

        for i in range(len(value_tp)):
            left = var_tp[i]
            right = value_tp[i]
            left.type = self.do_node(right)
            
        #raise UnsupportedException("Unsupported assignment syntax!")
            
    def do_Num(self, N: ast.Num):
#        dump(N)
        if isinstance(N.n, int):
            N.type = mytypes.int32
        elif isinstance(N.n, float):
            N.type = mytypes.float64
        else:
            raise UnhandledNodeException(N)
 #       print(N.type)
        
    def do_name_constant(self, N: ast.Num):
        if isinstance(N.value, bool):
            N.type = mytypes.bool
        else:
            raise UnhandledNodeException(N)

        
    def do_Tuple(self, N: ast.Tuple):
        types = []
        for ele in N.elts:            
            self.do_node(ele)
            types.append(ele.type)
        N.type = mytypes.TupleType(types)
        N.type.set_init_size(len(N.elts))

        
    def do_List(self, N: ast.List):
        types = []
        for ele in N.elts:            
            self.do_node(ele)
            types.append(ele.type)
            break        
        N.type = mytypes.ListType(types[0])
        N.type.set_init_size(len(N.elts))

        
    def annotate(self, node, type):
        node.type = type

        
    def do_For(self, N: ast.For):
        # Fields:
        # for 'target' in 'iter':
        #   'body'
        if N.type_comment == 'pfor':
            N.is_pfor = True
            self.cur_func.has_pfor = True
        else:
            N.is_pfor = False
            
        self.do_node(N.iter)
        container_ty = N.iter.type
        #N.target.type = N.iter.type.etype

        N.target.type = N.iter.type.get_elt_type()
        self.record_type(N.target)
        for expr in N.body:
            self.do_node(expr)

    def do_While(self, N: ast.While):
        for expr in N.body:
            self.do_node(expr)
            
    def do_If(self, N: ast.If):
        '''        
        if 'test':
          'body'
        'orelse':

        '''
        for expr in N.body:
            self.do_node(expr)
        
        if N.orelse:
            for expr in N.orelse:
                self.do_node(expr)
        
    def do_Return(self, N: ast.Return):
        # Return node doesn't need a type
        # TODO: could do a type check for return
        
        self.do_node(N.value)
        if self.cur_func.ret_type == mytypes.void:
            #assert False, "function return type unannotated!"
            self.cur_func.ret_type = N.value.type
        else:
            if not self.cur_func.ret_type.equals(N.value.type):
                m = 'Unmatched return types. \nPreviously: ' + str(self.cur_func.ret_type) + '\n'
                m += 'Now: ' + str(N.value.type)
                print(self.typemap)
                glb.exit_on_node(N, m)
        
        
    def visit_ImportFrom(self, node):
        for alias in node.names:
            self.stats["from"].append(alias.name)
        self.generic_visit(node)

    def report(self):
        pprint(self.stats)




from enum import Enum
import ast
from . import libfuncs
from . import mytypes
from . import defuse
from . import arrayopt

'''
Loop Invariant Code Motion for AST-level expressions (general)
'''

### Options ###
# 0: Nothing, 1: Warning, 2: Debug (light), 3: Debug (moderate), 4: Debug (heavy)
verbose_level = 1

dumpfunc = ast.dump


############################################################################################

class LicmAstOptimization(defuse.DefUseAnalysis):
    def __init__(self):
        defuse.DefUseAnalysis.__init__(self)

    def contains_nonpure_func(self, expr):
        if (isinstance(expr, ast.Call) and isinstance(expr.func, ast.Name)
            and expr.func.id not in defuse.pure_funcs):
            return True

        for node in ast.iter_child_nodes(expr):
            if self.contains_nonpure_func(node):
                return True

        return False

    def is_invstmt_candidate(self, node):
        return (node.kind is self.LevelTreeNode.Kind.ASSIGN
                or node.kind is self.LevelTreeNode.Kind.ANN_ASSIGN)

    def is_loop_init(self, node):
        return (node.kind is self.LevelTreeNode.Kind.FOR_IN
                or node.kind is self.LevelTreeNode.Kind.WHILE_IN)

    def collect_invariant_statements(self, loop_node):
        invstmt_nodes = []
        vars_ro = (loop_node.vars_use - loop_node.vars_def - loop_node.vars_cond_def
                   - loop_node.vars_may_def - loop_node.vars_em_def)
        var_i = loop_node.child.vars_def
        vars_l = (loop_node.vars_cond_def - loop_node.vars_use - var_i)
        if verbose_level >= 2:
            print('[Verbose] Search invariant statements in:', loop_node.kind, loop_node.level)
            if verbose_level >= 3:
                print('[Verbose] read-only vars:', vars_ro)
                print('[Verbose] loop index:', var_i)
                print('[Verbose] local vars:', vars_l)

        node = loop_node.child
        while node:
            if (node not in invstmt_nodes and self.is_invstmt_candidate(node)
                and node.vars_def.issubset(vars_l) and node.vars_use.issubset(vars_ro)):
                group = [node]
                used = False
                succ = node.next
                while succ:
                    def_succ = (succ.vars_def | succ.vars_cond_def
                                | succ.vars_may_def | succ.vars_em_def)
                    if def_succ & node.vars_def:
                        if (self.is_invstmt_candidate(succ)
                            and def_succ.issubset(node.vars_def)
                            and not used
                            and succ.vars_use.issubset(vars_ro | node.vars_def)):
                            group.append(succ)
                        else:
                            break
                    elif succ.vars_use & node.vars_def:
                        used = True
                    succ = succ.next

                if not succ:
                    for n in group:
                        if not self.contains_nonpure_func(n.ast_node):
                            if verbose_level >= 2:
                                print('[Verbose] Invariant statement found at:', n.level)
                                if verbose_level >= 3:
                                    n.show()
                                    if verbose_level >= 4: print(dumpfunc(n.ast_node))
                            invstmt_nodes.append(n)
                            vars_ro |= n.vars_def

            node = node.next

        return invstmt_nodes

    def verify_and_collect_cs_expr(self, expr: ast.expr, vars_ro: set, cs_exprs: dict, top = False):
        ope_nest = -1
        num_dims = -1

        if isinstance(expr, ast.Num):
            ope_nest = 0
            num_dims = 0
        elif isinstance(expr, ast.Name):
            if expr.id in vars_ro:
                ope_nest = 0
            if expr.id in self.curr_func.typemap:
                ty = self.curr_func.typemap[expr.id]
                if hasattr(ty, 'ndim'): num_dims = ty.ndim
                else: num_dims = 0
        elif isinstance(expr, ast.UnaryOp):
            n, d = self.verify_and_collect_cs_expr(expr.operand, vars_ro, cs_exprs)
            if n >= 0:
                ope_nest = n + 1
            num_dims = d
        else:
            valid = False
            calc_max = True
            calc_matmult = False
            arg1_minus1 = False
            ch = []
            if isinstance(expr, ast.BinOp):
                valid = True
                ch.append(expr.left)
                ch.append(expr.right)
            elif isinstance(expr, ast.BoolOp):
                valid = True
                ch.extend(expr.values)
            elif isinstance(expr, ast.Compare):
                valid = True
                ch.append(expr.left)
                ch.extend(expr.comparators)
            elif isinstance(expr, ast.Call):
                pure = arrayopt.is_call_to(expr, defuse.pure_funcs)
                valid = pure and not top
                calc_max = False
                if pure:
                    d = defuse.pure_funcs[expr.func.id]
                    if isinstance(d, int):
                        num_dims = d
                    elif d == 'arg1_len':
                        if isinstance(expr.args[0], list): num_dims = len(expr.args[0])
                        else: num_dims = 1
                    elif d == 'arg_max_dim':
                        calc_max = True
                    elif d == 'red_sum':
                        if len(expr.args) == 1: num_dims = 0
                        else: arg1_minus1 = True
                    elif d == 'matmult':
                        calc_matmult = True
                ch.extend(expr.args)
            elif isinstance(expr, ast.Subscript) and isinstance(expr.slice, ast.Index):
                valid = True
                calc_max = False
                index = expr.slice
                if isinstance(expr.value, ast.Name) and expr.value.id in self.curr_func.typemap:
                    ty = self.curr_func.typemap[expr.value.id]
                    ln = len(index.elts) if isinstance(index, ast.Tuple) else 1
                    if hasattr(ty, 'ndim') and ty.ndim >= ln:
                        num_dims = ty.ndim - ln
                ch.append(expr.value)
                if isinstance(index, ast.Tuple): ch.extend(index.elts)
                else: ch.append(index)

            if ch:
                num = len(ch)
                n = list(range(num))
                d = list(range(num))
                for i in range(num):
                    n[i], d[i] = self.verify_and_collect_cs_expr(ch[i], vars_ro, cs_exprs)

                if valid and min(n) >= 0:
                    ope_nest = sum(n) + 1
                else:
                    for i in range(num):
                        self.check_and_update_cs_expr(ch[i], n[i], d[i], cs_exprs)

                if calc_max and min(d) >= 0:
                    num_dims = max(d)
                elif arg1_minus1 and d[0] >= 1:
                    num_dims = d[0] - 1
                    if num_dims == 0: num_dims = 1
                elif calc_matmult and d[0] >= 1 and d[1] >= 1:
                    num_dims = d[0] + d[1] - 2
                    # 2+2: matmat, 1+2: vectmat, 2+1: matvect, 1+1: innerprod
            elif valid:
                ope_nest = 0
                num_dims = 0

        return ope_nest, num_dims

    def check_and_update_cs_expr(self, expr: ast.expr,
                                 ope_nest: int, num_dims: int, cs_exprs: dict):
        if ope_nest <= 0:
            return

        ty = None
        if hasattr(expr, 'type'):
            ty = expr.type
        elif isinstance(expr, ast.Call) and isinstance(expr.func, ast.Name):
            dummy = [mytypes.float64_ndarray, mytypes.float64_ndarray]
            ty = libfuncs.get_type(expr.func.id, 'pydd', dummy)

        if num_dims == -1 or not ty:
            if verbose_level >= 2:
                print('[Verbose] Invalid due to dimension (%s) or type (%s).' %(num_dims, ty))
                if verbose_level >= 4: print(dumpfunc(expr))
            return

        if isinstance(ty, mytypes.NpArray):
            assert num_dims >= 1
            ty = mytypes.NpArray(ty.etype, num_dims)
        cs_exprs[expr] = arrayopt.new_var(ty, self.curr_func)

        if verbose_level >= 2:
            print('[Verbose] Invariant expresson (nest-level = %s) found.' %ope_nest)
            if verbose_level >= 3:
                print('[Verbose] type =', ty, 'with ndims =', num_dims)
                if verbose_level >= 4: print(dumpfunc(expr))

    def collect_common_sub_expressions(self, loop_node, invstmt_nodes: list):
        cs_exprs = {}
        vars_ro = (loop_node.vars_use - loop_node.vars_def - loop_node.vars_cond_def
                   - loop_node.vars_may_def - loop_node.vars_em_def)
        for n in invstmt_nodes:
            vars_ro |= n.vars_def
        if verbose_level >= 2:
            print('[Verbose] Search invariant expressions in:', loop_node.kind, loop_node.level)
            if verbose_level >= 3:
                print('[Verbose] Read-only vars:', vars_ro)

        node = loop_node.child
        while node:
            if node not in invstmt_nodes:
                astn = None
                if node.is_assign() or node.kind is self.LevelTreeNode.Kind.CALL:
                    assert hasattr(node.ast_node, 'value')
                    astn = node.ast_node.value
                if astn:
                    b4size = len(cs_exprs)
                    n, d = self.verify_and_collect_cs_expr(astn, vars_ro, cs_exprs, True)
                    self.check_and_update_cs_expr(astn, n, d, cs_exprs)
                    if len(cs_exprs) > b4size:
                        node.has_cse = True

            node = node.next

        return cs_exprs

    def replace_cse_by_var(self, expr, cs_exprs: dict):
        if isinstance(expr, ast.stmt) and expr.value in cs_exprs:
            var = cs_exprs[expr.value]
            expr.value = arrayopt.new_name(var, ast.Load, self.curr_func.typemap[var])
        elif isinstance(expr, ast.UnaryOp) and expr.operand in cs_exprs:
            var = cs_exprs[expr.operand]
            expr.operand = arrayopt.new_name(var, ast.Load, self.curr_func.typemap[var])
        elif isinstance(expr, ast.BinOp):
            if expr.left in cs_exprs:
                var = cs_exprs[expr.left]
                expr.left = arrayopt.new_name(var, ast.Load, self.curr_func.typemap[var])
            if expr.right in cs_exprs:
                var = cs_exprs[expr.right]
                expr.right = arrayopt.new_name(var, ast.Load, self.curr_func.typemap[var])
        elif isinstance(expr, ast.BoolOp):
            n = len(expr.values)
            for i in range(n):
                if expr.values[i] in cs_exprs:
                    var = cs_exprs[expr.values[i]]
                    expr.values[i] = arrayopt.new_name(var, ast.Load,
                                                       self.curr_func.typemap[var])
        elif isinstance(expr, ast.Compare):
            if expr.left in cs_exprs:
                var = cs_exprs[expr.left]
                expr.left = arrayopt.new_name(var, ast.Load, self.curr_func.typemap[var])
            n = len(expr.comparators)
            for i in range(n):
                if expr.comparators[i] in cs_exprs:
                    var = cs_exprs[expr.comparators[i]]
                    expr.comparators[i] = arrayopt.new_name(var, ast.Load,
                                                            self.curr_func.typemap[var])
        elif isinstance(expr, ast.Call):
            n = len(expr.args)
            for i in range(n):
                if expr.args[i] in cs_exprs:
                    var = cs_exprs[expr.args[i]]
                    expr.args[i] = arrayopt.new_name(var, ast.Load,
                                                     self.curr_func.typemap[var])
        elif isinstance(expr, ast.Subscript) and isinstance(expr.slice, ast.Index):
            if expr.value in cs_exprs:
                var = cs_exprs[expr.value]
                expr.value = arrayopt.new_name(var, ast.Load, self.curr_func.typemap[var])
            index = expr.slice
            if isinstance(index, ast.Tuple):
                n = len(index.elts)
                for i in range(n):
                    if index.elts[i] in cs_exprs:
                        var = cs_exprs[index.elts[i]]
                        index.elts[i] = arrayopt.new_name(var, ast.Load,
                                                          self.curr_func.typemap[var])
            elif index in cs_exprs:
                var = cs_exprs[index]
                expr.slice = arrayopt.new_name(var, ast.Load, self.curr_func.typemap[var])

        for node in ast.iter_child_nodes(expr):
            self.replace_cse_by_var(node, cs_exprs)

    def licm_codegen(self, loop_node, loop_node0, invstmt_nodes: list, cs_exprs: dict):
        assert tuple(self.status) == loop_node.level and self.level_tree_current is loop_node
        assert loop_node.prev and not loop_node.parent
        self.level_tree_current = loop_node.prev
        self.status[-1] -= 1

        children = []
        ch = loop_node.child
        while ch:
            if ch not in invstmt_nodes: children.append(ch)
            ch = ch.next

        stmts = []
        for inv in invstmt_nodes:
            self.status[-1] += 1
            inv.level = tuple(self.status)
            self.level_tree_current.next = inv
            inv.prev = self.level_tree_current
            inv.parent = None
            inv.next = None
            inv.child = None
            self.level_tree_current = inv
            stmts.append(inv.ast_node)

        for inv in cs_exprs:
            tar = cs_exprs[inv]
            ty = self.curr_func.typemap[tar]
            stmt = arrayopt.new_assign(tar, inv, ty)
            self.visit_Assign(stmt)
            stmts.append(stmt)

        self.status[-1] += 1
        loop_node0.level = tuple(self.status)
        self.level_tree_current.next = loop_node0
        loop_node0.prev = self.level_tree_current
        assert not loop_node0.parent and not loop_node0.next
        self.level_tree_current = loop_node0

        body = []
        for ch in children:
            if self.is_loop_init(ch):
                self.status.append(0)
                ch.level = tuple(self.status)
                self.level_tree_current.child = ch
                ch.parent = self.level_tree_current
                assert not ch.prev and not ch.child
                ch.next = None
                self.level_tree_current = ch
            else:
                stmt = ch.ast_node
                if hasattr(ch, 'has_cse') and ch.has_cse:
                    self.replace_cse_by_var(stmt, cs_exprs)
                    if ch.kind is self.LevelTreeNode.Kind.ASSIGN:
                        self.visit_Assign(stmt)
                    elif ch.kind is self.LevelTreeNode.Kind.AUG_ASSIGN:
                        self.visit_AugAssign(stmt)
                    elif ch.kind is self.LevelTreeNode.Kind.ANN_ASSIGN:
                        self.visit_AnnAssign(stmt)
                    else:
                        self.visit_Call(stmt, True)
                else:
                    self.status[-1] += 1
                    ch.level = tuple(self.status)
                    self.level_tree_current.next = ch
                    ch.prev = self.level_tree_current
                    assert not ch.parent
                    ch.update_child_level()
                    ch.next = None
                    self.level_tree_current = ch
                if ch.kind is not self.LevelTreeNode.Kind.ELSE:
                    body.append(stmt)

        self.collect_child_def_use(loop_node0)
        loop_node0.ast_node.body = body
        stmts.append(loop_node0.ast_node)
        return stmts


############################################################################################

    def visit_FunctionDef(self, N: ast.FunctionDef):
        if verbose_level >= 2: print('[Verbose] LICM ast for function:', N.name)
        defuse.DefUseAnalysis.visit_FunctionDef(self, N)

    def visit_body(self, stmts: list):
        new_stmts = []
        transformed = False
        for stmt in stmts:
            ss = []
            if isinstance(stmt, ast.For):
                ss = self.visit_For(stmt)
            elif isinstance(stmt, ast.While):
                ss = self.visit_While(stmt)
            if ss:
                if len(ss) >= 2:
                    transformed = True
                new_stmts.extend(ss)
            else:
                self.visit_body_core(stmt)
                new_stmts.append(stmt)

        if transformed:
            stmts.clear()
            stmts.extend(new_stmts)

    def visit_For(self, N: ast.For):
        for_node = self.visit_For_head(N)
        for_copy = for_node.copy()

        self.visit_body(N.body)
        self.collect_child_def_use(for_node)

        invstmt_nodes = self.collect_invariant_statements(for_node)
        cs_exprs = self.collect_common_sub_expressions(for_node, invstmt_nodes)
        stmts = self.licm_codegen(for_node, for_copy, invstmt_nodes, cs_exprs)

        assert not N.orelse

        return stmts

    def visit_While(self, N: ast.While):
        while_node = self.visit_While_head(N)
        while_copy = while_node.copy()

        self.visit_body(N.body)
        self.collect_child_def_use(while_node)

        invstmt_nodes = self.collect_invariant_statements(while_node)
        cs_exprs = self.collect_common_sub_expressions(while_node, invstmt_nodes)
        stmts = self.licm_codegen(while_node, while_copy, invstmt_nodes, cs_exprs)

        assert not N.orelse

        return stmts



from enum import Enum
import ast
from . import libfuncs
from . import defuse

'''
Analysis for Loop Invariant Code motion (focuses on array references as first step).

Identify outermost location where array base address & shape are invariant.
Thereby it is legal to move the base address & shape computation to the location.

Example:
    while t < times:        # x2's base & shape is invariant within this while loop.
        x1 = zeros(n, float64())
        for i in range(n):  # x1's base & shape is invariant within this for loop.
            x1[i] = x2[i,i]
        ...

At the AST-level, each loop (i.e., ast.For or ast.While) may have attribute
"inv_subscripts", a list of invariant references (i.e., ast.Subscript).

Todo:
 - If same array is used multiple times, it tries to combine the location if possible
 - Standardize options, print (verbose and warning), and assert
'''

### Options ###
# 0: Nothing, 1: Debug (light), 2: Debug (modelate), 3: Debug (heavy)
verbose_level = 0

### Functions to access/modify ast.For/While nodes ###
def set_inv_subscripts(N, subscripts):
    N.inv_subscripts = subscripts

    global verbose_level
    if verbose_level >= 1:
        print('[Verbose] Loop at line', N.lineno, 'has invariant array subscripts:', end=' ')
        for subscr in subscripts:
            print(subscr.value.id, end=', ')
        print()

def set_inv_array_dim(N: ast.FunctionDef, inv_array_dim):
    N.inv_array_dim = inv_array_dim

    global verbose_level
    if verbose_level >= 1:
        print('[Verbose] Max dimensionality', inv_array_dim)

### Utilities ###
def get_num_array_dim(N: ast.Subscript):
    if isinstance(N.slice, ast.Index):
        if isinstance(N.slice, ast.Tuple):
            return len(N.slice.elts)
        else:
            return 1
    else:
        return -1


############################################################################################

class LicmArrayAnalysis(defuse.DefUseAnalysis):

    def __init__(self):
        '''
        Information per Function Definition
        '''
        defuse.DefUseAnalysis.__init__(self)
        self.subscr_to_node = {}   # Key: subscript, Value: LevelTreeNode where subscript is invariant

    def verify_fields(self):
        defuse.DefUseAnalysis.verify_fields(self)
        assert not self.subscr_to_node

    def clear_level_tree_information(self):
        defuse.DefUseAnalysis.clear_level_tree_information(self)
        self.subscr_to_node.clear()

    def add_subscr_to_node(self, arr: ast.Subscript):
        self.subscr_to_node[arr] = self.level_tree_current

    def compute_outermost_invariant_loop(self, arr_id, node):
        inv_loop = None
        curr = node
        while curr:
            if curr.prev:
                curr = curr.prev
            else:
                curr = curr.parent
                if curr:
                    if (arr_id in curr.vars_def
                        or arr_id in curr.vars_may_def
                        or arr_id in curr.vars_cond_def):
                        break
                    elif (curr.kind is self.LevelTreeNode.Kind.FOR
                          or curr.kind is self.LevelTreeNode.Kind.WHILE):
                        inv_loop = curr
        return inv_loop

    def eval_invariant_loop(self, N: ast.FunctionDef):
        global verbose_level
        if verbose_level >= 2:
            self.show_fields(verbose_level >= 3)

        node_to_subscrs = {}  # Key: LevelTreeNode, Value: subscripts
        inv_array_dim = {}    # Key: array (ID), Value: max dimensionality
        for subscr in self.subscr_to_node:
            if isinstance(subscr.value, ast.Name):
                array_id = subscr.value.id
                inv_loop = self.compute_outermost_invariant_loop(array_id, self.subscr_to_node[subscr])
                ndim = get_num_array_dim(subscr)
                if inv_loop and ndim > 0:
                    self.subscr_to_node[subscr] = inv_loop
                    if inv_loop in node_to_subscrs:
                        node_to_subscrs[inv_loop].append(subscr)
                    else:
                        node_to_subscrs[inv_loop] = [subscr]

                    if array_id in inv_array_dim:
                        if ndim > inv_array_dim[array_id]:
                            inv_array_dim[array_id] = ndim
                    else:
                        inv_array_dim[array_id] = ndim

        for node in node_to_subscrs:
            subscripts = node_to_subscrs[node]
            set_inv_subscripts(node.ast_node, subscripts)

        set_inv_array_dim(N, inv_array_dim)


############################################################################################

    def visit_FunctionDef(self, N: ast.FunctionDef):
        self.verify_fields()

        self.visit_FunctionDef_main(N)

        self.eval_invariant_loop(N)

        self.clear_level_tree_information()

    def visit_Subscript(self, N: ast.Subscript):
        self.add_subscr_to_node(N)
        self.visit(N.value)
        self.visit(N.slice)



from . import glb
from .symboltable import symtab
from .glb import dump
from . import mytypes
from . import utils
import ast

class Scanner(ast.NodeVisitor):
    def __init__(self):
        self.cur_func = None
    
    def visit_Module(self, N):
        self.generic_visit(N)

    def visit_Import(self, N):        
        glb.cpp_module.add_imported_module(N.names[0])
        
    def visit_FunctionDef(self, N: ast.FunctionDef):
        N.ret_type = mytypes.void
        if N.returns:
            N.ret_type = utils.get_annotation_type(N.returns)
        N.localtypes = {} 
        symtab.register_user_func(N)
        
        
